from __future__ import division

import base64
import json
import logging
import os

import requests

from ._base import BaseClient
from ..bencode import bencode

logger = logging.getLogger(__name__)

class UnableToLoginException(Exception):
    pass

class RPCCallFailedException(Exception):
    pass

class TransmissionVersionTooLowException(Exception):
    pass

class TransmissionClient(BaseClient):
    identifier = 'transmission'

    _session_id = ''
    def __init__(self, url):
        """
        Initializes a new Transmission client.

        url - The url where transmission rpc can be reached.
        """
        self.url = url

    def _call(self, method, **kwargs):
        """
        Actual calls Transmission JSON RPC.
        """
        logger.debug('Calling %r args %r' % (method, kwargs))
        return requests.post(self.url, data=json.dumps({'method': method, 'arguments': kwargs}), headers={'X-Transmission-Session-Id': self._session_id})

    def call(self, method, **kwargs):
        """
        Calls Transmission JSON RPC.
        """
        r = self._call(method, **kwargs)
        if r.status_code == 409:
            self._session_id = r.headers['X-Transmission-Session-Id']
            r = self._call(method, **kwargs)

        if r.status_code != 200:
            raise UnableToLoginException()

        r = r.json()
        logger.debug('Got transmission reply %r' % r)
        if r['result'] != 'success':
            raise RPCCallFailedException()

        return r['arguments']

    def get_config(self):
        """
        Get the current configuration that can be used in the autotorrent config file
        """
        return {
            'url': self.url,
        }

    @classmethod
    def auto_config(cls):
        """
        Tries to auto configure transmission using the .config/transmission-daemon/settings.json file
        """
        config_path = os.path.expanduser('~/.config/transmission-daemon/settings.json')
        if not os.path.isfile(config_path):
            logger.debug('transmission daemon config file was not found')
            return

        if not os.access(config_path, os.R_OK):
            logger.debug('Unable to access transmission daemon config file at %s' % config_path)
            return

        with open(config_path, 'r') as f:
            config_data = json.load(f)

        ip = config_data.get('rpc-bind-address')
        port = config_data.get('rpc-port')
        if ip == '0.0.0.0':
            ip = '127.0.0.1'

        if not ip:
            logger.debug('Unable to find a bind ip')
            return

        if not port:
            logger.debug('Unable to find port')
            return

        return cls('http://%s:%s/transmission/rpc' % (ip, port))

    def test_connection(self):
        """
        Tests the Transmission JSONRPC connection, returns message if found.
        """
        session_data = self.call('session-get')
        if session_data['rpc-version'] < 15:
            raise TransmissionVersionTooLowException('You need to update to a newer version of Transmission')

        return '%s, config-dir: %s, download-dir: %s' % (session_data['version'], session_data['config-dir'], session_data['download-dir'])

    def get_torrents(self):
        """
        Returns a set of info hashes currently added to the client.
        """
        logger.info('Getting a list of torrent hashes')
        result = self.call('torrent-get', fields=['hashString'])
        return set(x['hashString'].lower() for x in result['torrents'])

    def add_torrent(self, torrent, destination_path, files, fast_resume=True):
        """
        Add a new torrent to Transmission.

        torrent is the decoded file as a python object.
        destination_path is where the links are. The complete files must be linked already.
        files is a list of files found in the torrent.
        """
        name = torrent[b'info'][b'name']
        logger.info('Trying to add a new torrent to transmission: %r' % name)

        destination_path = os.path.abspath(destination_path)

        encoded_torrent = base64.b64encode(bencode(torrent))

        if b'files' in torrent[b'info']:
            download_dir = os.path.dirname(destination_path)
            display_name = os.path.basename(destination_path)
        else:
            download_dir = destination_path
            display_name = name.decode('utf-8')

        kwargs = {'download-dir': download_dir, 'metainfo': encoded_torrent.decode('utf-8'), 'paused': True}
        result = self.call('torrent-add', **kwargs)
        tid = result['torrent-added']['id']

        self.call('torrent-rename-path', ids=[tid], path=name.decode('utf-8'), name=display_name)
        self.call('torrent-start', ids=[tid])

        return True



import logging

logger = logging.getLogger(__name__)

TORRENT_CLIENTS = {}

try:
    from .rtorrent import RTorrentClient
    TORRENT_CLIENTS[RTorrentClient.identifier] = RTorrentClient
    
    logger.debug('Enabled client %s' % RTorrentClient.identifier)
except ImportError:
    logger.debug('Failed to enable client rtorrent')

try:
    from .deluge import DelugeClient
    TORRENT_CLIENTS[DelugeClient.identifier] = DelugeClient
    
    logger.debug('Enabled client %s' % DelugeClient.identifier)
except ImportError:
    logger.debug('Failed to enable client deluge')

try:
    from .transmission import TransmissionClient
    TORRENT_CLIENTS[TransmissionClient.identifier] = TransmissionClient
    
    logger.debug('Enabled client %s' % TransmissionClient.identifier)
except ImportError:
    logger.debug('Failed to enable client transmission')


from __future__ import division

import hashlib
import logging
import os
import re

__all__ = [
    'is_unsplitable',
    'get_root_of_unsplitable',
    'Pieces',
]

UNSPLITABLE_FILE_EXTENSIONS = [
    set(['.rar', '.sfv']),
    set(['.mp3', '.sfv']),
    set(['.vob', '.ifo']),
]

logger = logging.getLogger(__name__)

def is_unsplitable(files):
    """
    Checks if a list of files can be considered unsplitable, e.g. VOB/IFO or scene release.
    This means the files can only be used in this combination.
    """
    extensions = set(os.path.splitext(f)[1].lower() for f in files)
    found_unsplitable_extensions = False
    for exts in UNSPLITABLE_FILE_EXTENSIONS:
        if len(extensions & exts) == len(exts):
            found_unsplitable_extensions = True
            break
    
    lowercased_files = set([f.lower() for f in files])
    found_magic_file = False
    if 'movieobject.bdmv' in lowercased_files:
        found_magic_file = True
    
    return found_unsplitable_extensions or found_magic_file

def get_root_of_unsplitable(path):
    """
    Scans a path for the actual scene release name, e.g. skipping cd1 folders.
    
    Returns None if no scene folder could be found
    """
    path = path[::-1]
    for p in path:
        if not p:
            continue
        
        if re.match(r'^(cd[1-9])|(samples?)|(proofs?)|((vob)?sub(title)?s?)$', p, re.IGNORECASE): # scene paths
            continue
        
        if re.match(r'^(bdmv)|(disc\d*)|(video_ts)$', p, re.IGNORECASE): # bluray / dd
            continue
        
        
        return p

class Pieces(object):
    """
    Can help check if files match the files found in a torrent.
    """
    
    def __init__(self, torrent):
        self.piece_size = torrent[b'info'][b'piece length']
        self.pieces = []
        for i in range(0, len(torrent[b'info'][b'pieces']), 20):
            self.pieces.append(torrent[b'info'][b'pieces'][i:i+20])
    
    def get_complete_pieces(self, start_size, end_size):
        """
        Finds complete pieces and returns the alignment needed from
        the beginning and the end (to match the file).
        """
        logger.debug('Getting complete pieces for file starting at %i and ending at %i. Piece size is %i' % (start_size, end_size, self.piece_size))
        
        start_piece, start_offset = divmod(start_size, self.piece_size)
        if start_offset:
            start_piece += 1
        
        if start_offset:
            start_offset = self.piece_size - start_offset
        
        end_piece, end_offset = divmod(end_size, self.piece_size)

        logger.debug('Start piece:%i end piece:%i' % (start_piece, end_piece-1))
        return start_offset, end_offset, self.pieces[start_piece:end_piece]
    
    def find_piece_breakpoint(self, file_path, start_size, end_size):
        """
        Finds the point where a file with a different size is modified and tries to align it with pieces.
        """
        start_offset, end_offset, pieces = self.get_complete_pieces(start_size, end_size)
        
        failed_pieces = (len(pieces) // 20) or 1 # number of pieces that can fail in a row and then put an end to checking
        success_count = failed_pieces
        piece_status = []
        
        with open(file_path, 'rb') as f:
            f.seek(start_offset)
            for i, piece in enumerate(pieces):
                logger.debug('Checking piece %i for breakingpoint' % (i, ))
                h = hashlib.sha1(f.read(self.piece_size)).digest()
                if h == piece:
                    logger.debug('Piece %i matched' % i)
                    if success_count < failed_pieces:
                        success_count += 1
                    piece_status.append(True)
                else:
                    logger.debug('Piece %i did not match' % i)
                    success_count -= 1
                    piece_status.append(False)
                
                if success_count <= 0:
                    logger.debug('The breakingpoint has been found after piece %i - more than %i failed pieces' % (i, failed_pieces))
                    break
        
        for p in piece_status[::-1]:
            if p:
                break
            i -= 1
        
        breakingpoint = start_offset + self.piece_size*i
        logger.debug('A total of %i pieces were ok, so we set breakingpoint at %i' % (i, breakingpoint))
        return breakingpoint

    def match_file(self, file_path, start_size, end_size):
        """
        Try to match file starting at start_size and ending at end_size.
        """
        start_offset, end_offset, pieces = self.get_complete_pieces(start_size, end_size)
        logger.debug('Stuff to check start_offset:%i end_offset:%i pieces:%s' % (start_offset, end_offset, len(pieces)))
        if not pieces:
            logger.debug('No whole pieces found for %r, taking this as a not-match' % file_path)
            return False, False
        
        check_pieces = (len(pieces) // 10) or 1
        
        match_start, match_end = 0, 0
        size = os.path.getsize(file_path)
        with open(file_path, 'rb') as f:
            for i in range(check_pieces): # check from beginning
                seek_offset = start_offset+self.piece_size*i
                logger.debug('Checking piece %i from beginning of file, reading from %i bytes. Filesize: %i' % (i, seek_offset, size))
                f.seek(seek_offset)
                h = hashlib.sha1(f.read(self.piece_size)).digest()
                logger.debug('Matching hash %r against %r' % (h, pieces[i]))
                if h == pieces[i]:
                    logger.debug('Piece %i matched' % i)
                    match_start += 1
                else:
                    logger.debug('Piece %i did not match' % i)
            
            for i in range(check_pieces): # check from end
                seek_offset = size-end_offset-self.piece_size*(i+1)
                logger.debug('Checking piece %i from end of file, reading from %i bytes. Filesize: %i' % (i, seek_offset, size))
                f.seek(seek_offset)
                h = hashlib.sha1(f.read(self.piece_size)).digest()
                piece = pieces[(i+1)*-1]
                logger.debug('Matching hash %r against %r' % (h, piece))
                if h == piece:
                    logger.debug('Piece %i matched' % i)
                    match_end += 1
                else:
                    logger.debug('Piece %i did not match' % i)
        
        logger.debug('Checked %i pieces from both start and end. %i matched from start and %i matched from end.' % (check_pieces, match_start, match_end))
        
        if check_pieces < 4:
            must_match = 1
        elif check_pieces < 10:
            must_match = 2
        else:
            must_match = max(check_pieces // 10, 3)
        
        return (match_start and check_pieces - match_start <= must_match,
                match_end and check_pieces - match_end <= must_match)



from __future__ import division

import base64
import hashlib
import logging
import os
import re

from deluge_client import DelugeRPCClient

from ._base import BaseClient
from ..bencode import bencode
from ..humanize import humanize_bytes

logger = logging.getLogger(__name__)

class UnableToLoginException(Exception):
    pass

class DelugeClient(BaseClient):
    identifier = 'deluge'

    def __init__(self, host, username, password):
        """
        Initializes a new Deluge client.

        url - The url where deluge json can be reached.
        password - The password used to login
        """
        host, port = host.split(':')
        self.host = host
        self.port = int(port)
        self.username = username
        self.password = password
        self.rpcclient = DelugeRPCClient(self.host, self.port, self.username, self.password, decode_utf8=True)

    def _login(self):
        """
        Logs into deluge
        """
        if not self.rpcclient.connected:
            self.rpcclient.connect()

    def get_config(self):
        """
        Get the current configuration that can be used in the autotorrent config file
        """
        return {
            'host': '%s:%s' % (self.host, self.port),
            'username': self.username,
            'password': self.password,
        }

    @classmethod
    def auto_config(cls):
        """
        Tries to auto configure deluge using the .config/deluge files
        """
        config_path = os.path.expanduser('~/.config/deluge/core.conf')
        if not os.path.isfile(config_path):
            logger.debug('deluge config file was not found')
            return

        if not os.access(config_path, os.R_OK):
            logger.debug('Unable to access deluge config file at %s' % config_path)
            return

        auth_path = os.path.expanduser('~/.config/deluge/auth')
        if not os.path.isfile(auth_path):
            logger.debug('deluge auth file was not found')
            return

        if not os.access(auth_path, os.R_OK):
            logger.debug('Unable to access deluge confauthig file at %s' % auth_path)
            return

        with open(config_path, 'r') as f:
            config_data = f.read()

        daemon_port = re.findall('"daemon_port":\s*(\d+)', config_data)
        if not daemon_port:
            logger.debug('No deluge port, just trying to use default')
            daemon_port = 58846
        else:
            daemon_port = int(daemon_port[0])

        with open(auth_path, 'r') as f:
            auth_data = f.read()

        auth_data = auth_data.split('\n')[0].split(':')
        if len(auth_data[0]) < 2:
            logger.debug('Invalid entry found in auth file')
            return

        username = auth_data[0]
        password = auth_data[1]

        return cls('127.0.0.1:%s' % daemon_port, username, password)

    def test_connection(self):
        """
        Tests the Deluge RPC connection, returns message if found.
        """
        self._login()
        return 'Free space: %s' % humanize_bytes(self.rpcclient.call('core.get_free_space'))

    def get_torrents(self):
        """
        Returns a set of info hashes currently added to the client.
        """
        logger.info('Getting a list of torrent hashes')
        self._login()
        result = self.rpcclient.call('core.get_torrents_status', {}, ['name'])
        return set(x.lower().decode('ascii') for x in result.keys())

    def add_torrent(self, torrent, destination_path, files, fast_resume=True):
        """
        Add a new torrent to Deluge.

        torrent is the decoded file as a python object.
        destination_path is where the links are. The complete files must be linked already.
        files is a list of files found in the torrent.
        """
        name = torrent[b'info'][b'name']
        logger.info('Trying to add a new torrent to deluge: %r' % name)

        destination_path = os.path.abspath(destination_path)

        infohash = hashlib.sha1(bencode(torrent[b'info'])).hexdigest()
        encoded_torrent = base64.b64encode(bencode(torrent))

        basename = os.path.basename(destination_path)
        mapped_files = {}
        for i, f in enumerate(files):
            mapped_files[i] = os.path.join(basename, *f['path'])

        self._login()
        result = self.rpcclient.call('core.add_torrent_file', 'torrent.torrent', encoded_torrent, {
                                                                'download_location': os.path.dirname(destination_path),
                                                                'mapped_files': mapped_files,
                                                                'seed_mode': fast_resume})

        return result and result.decode('utf-8') == infohash



from __future__ import division, unicode_literals

import hashlib
import logging
import os
import shelve

from fnmatch import fnmatch

from .utils import is_unsplitable, get_root_of_unsplitable

logger = logging.getLogger(__name__)

class Database(object):
    hash_mode_size_varying = 10.0 # 10% size variation from size on disk for the two scan modes
                                  # that allows size to vary
    
    def __init__(self, db_file, paths, ignore_files, normal_mode, unsplitable_mode, exact_mode,
                 hash_name_mode, hash_size_mode, hash_slow_mode):
        """
        Database used to match files and torrents.
        """
        self.db = shelve.open(db_file)
        self.db_file = db_file
        self.paths = paths
        self.ignore_files = [self.normalize_filename(x) for x in ignore_files]
        self.normal_mode = normal_mode
        self.unsplitable_mode = unsplitable_mode
        self.exact_mode = exact_mode
        self.hash_name_mode = hash_name_mode
        self.hash_size_mode = hash_size_mode
        self.hash_slow_mode = hash_slow_mode
        self.hash_mode = hash_name_mode or hash_size_mode or hash_slow_mode
        self.hash_size_table = None
    
    def truncate(self):
        """
        Truncates the database
        """
        logger.info('Truncated the database')
        self.db.close()
        self.db = shelve.open(self.db_file, flag='n')
    
    def insert_into_database(self, root, f, mode, prefix=None, unsplitable_name=None):
        """
        Wraps the database insert to catch exceptions
        """
        try:
            self._insert_into_database(root, f, mode, prefix, unsplitable_name)
        except UnicodeDecodeError:
            logger.error('Failed to insert %r / %r / %r' % (root, f, mode))

    def _insert_into_database(self, root, f, mode, prefix=None, unsplitable_name=None):
        """
        Does the actual insertion into the database.
        """
        path = os.path.abspath(os.path.join(root, f))
        if not os.access(path, os.R_OK):
            logger.warning('Path %r is not accessible, skipping' % path)
            return
        
        if mode == 'exact':
            key = self.keyify(prefix, f)
            if key in self.db:
                self.db[key] = self.db[key] + [path]
            else:
                self.db[key] = [path]
        else:
            normalized_filename = self.normalize_filename(f)
            size = os.path.getsize(path)
            
            if mode.startswith('hash_'):
                if mode == 'hash_store_name': # the size can vary, name is exact. I.e. filename to path mapping
                    key = self.keyify(normalized_filename)
                elif mode == 'hash_store_size': # the name can vary, size is exact (same db can be used for slow-mo). I.e. size to path mapping
                    key = str('s:%i' % size)
                self.db[key] = self.db.get(key, []) + [path]
            else:
                if mode == 'unsplitable':
                    split_root = root.split(os.sep)
                    p_index = len(split_root) - split_root[::-1].index(unsplitable_name) - 1
                    p = [self.normalize_filename(x) for x in split_root[p_index:]] + [normalized_filename]
                    
                    key = self.keyify(size, *p)
                elif mode == 'normal':
                    key = self.keyify(size, normalized_filename)
                
                if key in self.db: # check if same file
                    old_inode = os.stat(self.db[key]).st_ino
                    new_inode = os.stat(path).st_ino
                    if old_inode != new_inode:
                        logger.warning('Duplicate key %s and %s' % (path, self.db[key]))
    
                self.db[key] = path
    
    def skip_file(self, f):
        """
        Checks if a filename is in the skiplist
        """
        normalized_filename = self.normalize_filename(f)
        for ignore_file in self.ignore_files:
            if fnmatch(normalized_filename, ignore_file):
                return True
        return False
    
    def rebuild(self, paths=None):
        """
        Scans the paths for files and rebuilds the database.
        """
        if paths:
            logger.info('Just adding new paths')
        else:
            logger.info('Rebuilding database')
            self.truncate()
            paths = self.paths
        
        unsplitable_paths = set()
        if self.unsplitable_mode or self.exact_mode:
            logger.info('Special modes enabled, doing a preliminary scan')
            for root_path in paths:
                logger.info('Preliminary scanning %s' % root_path)
                for root, dirs, files in os.walk(root_path):
                    if is_unsplitable(files):
                        sep_root = root.split(os.sep)
                        name = get_root_of_unsplitable(root.split(os.sep))
                        while sep_root[-1] != name:
                            sep_root.pop()
                        path = os.path.join(*sep_root)
                        logger.debug('Found unsplitable path %r' % path)
                        unsplitable_paths.add(path)
                logger.info('Done preliminary scanning %s' % root_path)
        
        for root_path in paths:
            logger.info('Scanning %s' % root_path)
            for root, dirs, files in os.walk(root_path):
                unsplitable = False
                if self.unsplitable_mode or self.exact_mode:
                    sep_root = root.split(os.sep)
                    while sep_root:
                        if os.path.join(*sep_root) in unsplitable_paths:
                            break
                        sep_root.pop()
                    
                    if sep_root:
                        unsplitable = True
                        if self.unsplitable_mode:
                            unsplitable_name = sep_root[-1]
                            logger.info('Looks like we found a unsplitable release in %r' % (os.sep.join(sep_root)))
                            for f in files:
                                self.insert_into_database(root, f, 'unsplitable', unsplitable_name=unsplitable_name)
                            continue
                
                if not unsplitable:
                    if self.normal_mode:
                        for f in files:
                            if self.skip_file(f):
                                continue
                            
                            self.insert_into_database(root, f, 'normal')
                        
                    if self.exact_mode:
                        for f in files:
                            self.insert_into_database(root, f, 'exact', 'f')
                        
                        for d in dirs:
                            self.insert_into_database(root, d, 'exact', 'd')
                
                if self.hash_name_mode or self.hash_size_mode or self.hash_slow_mode:
                    for f in files:
                        if self.hash_size_mode or self.hash_slow_mode:
                            self.insert_into_database(root, f, 'hash_store_size')
                        
                        if self.hash_name_mode:
                            self.insert_into_database(root, f, 'hash_store_name')

                    
            logger.info('Done scanning %s' % root_path)
        self.db.sync()
    
    def clear_hash_size_table(self):
        """
        Clears the hash size table.
        """
        self.hash_size_table = None
    
    def build_hash_size_table(self):
        """
        Builds a table of all sizes to make lookups faster for varying sizes.
        """
        if self.hash_size_table is not None:
            logger.debug('Hash size table already built, skipping')
            return
        
        self.hash_size_table = set()
        for key in self.db.keys():
            if not key.startswith('s:'):
                continue
            
            _, size = key.split(':')
            self.hash_size_table.add(int(size))
        
        self.hash_size_table = sorted(self.hash_size_table)
    
    def find_hash_varying_size(self, size):
        """
        Looks for a file with close to size in the database.
        The function assumes build_hash_size_table has already been called.
        
        Returns a list of paths ordered by how close they are to the size.
        """
        size_span = size * self.hash_mode_size_varying / 100
        min_size_span, max_size_span = size - size_span, size + size_span
        
        found_sizes = []
        for db_size in self.hash_size_table:
            if db_size < min_size_span:
                continue
            
            if db_size > max_size_span:
                break
            
            found_sizes.append(db_size)
        
        found_sizes = sorted(found_sizes, key=lambda x:abs(x-size))
        result = []
        for found_size in found_sizes:
            key = str('s:%i' % found_size)
            result += self.db.get(key, [])
        
        return result
    
    def find_hash_size(self, size):
        """
        Looks for a file with exact size in the database.
        
        Returns a list of paths.
        """
        return self.db.get(str('s:%s' % size), [])
    
    def find_hash_name(self, f):
        """
        Looks for a file with name f in the database.
        
        Returns a list of paths.
        """
        key = self.keyify(self.normalize_filename(f))
        
        return self.db.get(key, [])
    
    def find_unsplitable_file_path(self, rls, f, size):
        """
        Looks for a file in the database.
        """
        f = [self.normalize_filename(x) for x in f]
        key = self.keyify(size, self.normalize_filename(rls), *f)

        return self.db.get(key)
    
    def find_exact_file_path(self, prefix, rls):
        """
        Looks for a name in the database.
        """
        key = self.keyify(prefix, rls)

        return self.db.get(key)
    
    def find_file_path(self, f, size):
        """
        Looks for a file in the database.
        """
        key = self.keyify(size, self.normalize_filename(f))

        return self.db.get(key)
    
    def keyify(self, size, *names):
        """
        Turns a name and size into a key that can be stored in the database.
        """
        key = '%s|%s' % (size, '|'.join(names))
        logger.debug('Keyify: %s' % key)
        
        return hashlib.sha256(key.encode('utf-8')).hexdigest()
    
    def normalize_filename(self, filename):
        """
        Normalizes a filename to better detect simlar files.
        """
        return filename.replace(' ', '_').lower()



class BaseClient(object):
    identifier = None # used to identify it in the config file
    
    @classmethod
    def auto_config(cls):
        """
        Tries to autoconfigure the client
        """
        raise NotImplemented
    
    def get_config(self):
        """
        Gets the current configuration as a dict that can
        be used to create an entry in the .conf file.
        """
        raise NotImplemented
    
    def test_connection(self):
        """
        Tests the connection to the torrent client and returns
        random information from the client.
        """
        raise NotImplemented
    
    def get_torrents(self):
        """
        Returns a list of torrents currently addd to the client.
        
        A set of ascii strings.
        """
        raise NotImplemented
    
    def add_torrent(self, torrent, destination_path, files, fast_resume=True):
        """
        Adds a torrent to the torrent client.
        """
        raise NotImplemented


import argparse
import json
import logging
import os
import shutil

from six.moves import configparser, input

from autotorrent.at import AutoTorrent
from autotorrent.clients import TORRENT_CLIENTS
from autotorrent.db import Database
from autotorrent.humanize import humanize_bytes

def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        print(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")
        

def commandline_handler():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", dest="config_file", default="autotorrent.conf", help="Path to config file")
    parser.add_argument("-l", "--client", dest="client", default="default", help="Name of client to use (when multiple configured)")
    parser.add_argument("--create_config", dest="create_config_file", nargs='?', const='autotorrent.conf', default=None, help="Creates a new configuration file")
    
    parser.add_argument("-t", "--test_connection", action="store_true", dest="test_connection", default=False, help='Tests the connection to the torrent client')
    parser.add_argument("--dry-run", nargs='?', const='txt', default=None, dest="dry_run", choices=['txt', 'json'], help="Don't do any actual adding, just scan for files needed for torrents.")
    parser.add_argument("-r", "--rebuild", dest="rebuild", default=False, help='Rebuild the database', nargs='*')
    parser.add_argument("-a", "--addfile", dest="addfile", default=False, help='Add a new torrent file to client', nargs='+')
    parser.add_argument("-d", "--delete_torrents", action="store_true", dest="delete_torrents", default=False, help='Delete torrents when they are added to the client')
    parser.add_argument("--verbose", help="increase output verbosity", action="store_true", dest="verbose")
    
    args = parser.parse_args()
    
    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.ERROR)
    
    if args.create_config_file: # autotorrent.conf
        if os.path.exists(args.create_config_file):
            parser.error("Target %r already exists, not creating" % args.create_config_file)
        else:
            src = os.path.join(os.path.dirname(__file__), 'autotorrent.conf.dist')
            shutil.copy(src, args.create_config_file)
            print('Created configuration file %r' % args.create_config_file)

        client_config = {
            'client': 'rtorrent',
            'url': 'http://user:pass@127.0.0.1/RPC2',
            'label': 'autotorrent',
        }
        
        if query_yes_no('Do you want to try and auto-configure torrent client?'):
            working_clients = []
            for client_name, cls in TORRENT_CLIENTS.items():
                obj = cls.auto_config()
                try:
                    if obj.test_connection():
                        working_clients.append(obj)
                except:
                    continue
            
            if working_clients:
                print('Found %i clients - please choose a client to use' % len(working_clients))
                for i, client in enumerate(working_clients, 1):
                    print('[%i] %s' % (i, client.identifier))
                print('[0] None of the above - do not auto-configure any client\n')
                
                while True:
                    error = False
                    try:
                        choice = int(input('> '))
                    except ValueError:
                        error = True
                    else:
                        if len(working_clients) < choice or choice < 0:
                            error = True
                    
                    if error:
                        print('Invalid choice, please choose again')
                    else:
                        if choice > 0:
                            client = working_clients[choice-1]
                            print('Setting client to %s' % client.identifier)
                            client_config = client.get_config()
                            client_config['client'] = client.identifier
                        
                        break
            else:
                print('Unable to auto-detect any clients, you will have to configure it manually.')
            
        config = configparser.ConfigParser()
        config.read(args.create_config_file)
        for k, v in client_config.items():
            config.set('client', k, v)
            
        with open(args.create_config_file, 'w') as configfile:
            config.write(configfile)
        
        quit()
    
    if not os.path.isfile(args.config_file):
        parser.error("Config file not found %r" % args.config_file)
    
    config = configparser.ConfigParser()
    config.read(args.config_file)
    
    current_path = os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(args.config_file))) # Changing directory to where the config file is.
    
    if not config.has_section('general'):
        parser.error('AutoTorrent is not properly configured, please edit %r' % args.config_file)
        quit(1)

    disks = [disk for _, disk in config.items('disks')]
    
    scan_mode = set(config.get('general', 'scan_mode').split(','))
    
    exact_mode = 'exact' in scan_mode
    unsplitable_mode = 'unsplitable' in scan_mode
    normal_mode = 'normal' in scan_mode

    hash_name_mode = 'hash_name' in scan_mode
    hash_size_mode = 'hash_size' in scan_mode
    hash_slow_mode = 'hash_slow' in scan_mode
    
    db = Database(config.get('general', 'db'), disks,
                  config.get('general', 'ignore_files').split(','),
                  normal_mode, unsplitable_mode, exact_mode,
                  hash_name_mode, hash_size_mode, hash_slow_mode)
    
    client_option = 'client'
    if args.client != 'default':
        client_option += '-%s' % args.client
    
    try:
        client_name = config.get(client_option, 'client')
    except configparser.NoSectionError:
        print('It seems like %r is not a configured client' % args.client)
        quit(1)
    
    if client_name not in TORRENT_CLIENTS:
        print('Unknown client %r - Known clients are: %s' % (client_name, ', '.join(TORRENT_CLIENTS.keys())))
        quit(1)
    
    client_options = dict(config.items(client_option))
    client_options.pop('client')
    client = TORRENT_CLIENTS[client_name](**client_options)
    
    at = AutoTorrent(
        db,
        client,
        config.get('general', 'store_path'),
        config.getint('general', 'add_limit_size'),
        config.getfloat('general', 'add_limit_percent'),
        args.delete_torrents,
        (config.get('general', 'link_type') if config.has_option('general', 'link_type') else 'soft'),
    )
    
    if args.test_connection:
        proxy_test_result = client.test_connection()
        if proxy_test_result:
            print('Connected to torrent client successfully!')
            print('  result: %s' % proxy_test_result)
    
    if isinstance(args.rebuild, list):
        if args.rebuild:
            print('Adding new folders to database')
            db.rebuild(args.rebuild)
            print('Added to database')
        else:
            print('Rebuilding database')
            db.rebuild()
            print('Database rebuilt')

    
    if args.addfile:
        dry_run = bool(args.dry_run)
        dry_run_data = []
        
        print('Found %s torrent(s)' % len(args.addfile))
        if not dry_run:
            at.populate_torrents_seeded()
        
        for torrent in args.addfile:
            result = at.handle_torrentfile(os.path.join(current_path, torrent), dry_run)
            if dry_run:
                dry_run_data.append({
                    'torrent': torrent,
                    'found_bytes': result[0],
                    'missing_bytes': result[1],
                    'would_add': not result[2],
                    'local_files': result[3],
                })
        
        if dry_run:
            if args.dry_run == 'json':
                print(json.dumps(dry_run_data))
            elif args.dry_run == 'txt':
                for torrent in dry_run_data:
                    print('Torrent: %s' % torrent['torrent'])
                    print(' Found data: %s - Missing data: %s - Would add: %s' % (humanize_bytes(torrent['found_bytes']),
                                                                                  humanize_bytes(torrent['missing_bytes']),
                                                                                  torrent['would_add'] and 'Yes' or 'No'))
                    print(' Local files used:')
                    for f in torrent['local_files']:
                        print('  %s' % f)
                    print('')

if __name__ == '__main__':
    commandline_handler()
